/*
 * Copyright (c) 2003-2005 The BISON Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package mycode.treefiledistr;

import peersim.config.*;
import peersim.core.*;
import peersim.transport.Transport;
import peersim.edsim.*;


/**
 * Initialize an aggregation protocol using a peak distribution; only one peak
 * is allowed. Note that any protocol implementing
 * {@link peersim.vector.SingleValue} can be initialized by this component.
 * 
 */
public class TFDInitializer implements Control {

    // ------------------------------------------------------------------------
    // Constants
    // ------------------------------------------------------------------------

    private static final String PAR_MINCHILDREN = "minchildren";
    private static final String PAR_MAXCHILDREN = "maxchildren";
    private static final String PAR_MAXLEVELS = "maxlevels";
    private static final String PAR_NUMCHUNKS = "numchunks";

    /**
     * The protocol to operate on.
     * 
     * @config
     */
    private static final String PAR_PROT = "protocol";

    private static final double[] inputBandwidth = {0.005, 0.05, 0.1 };
//    private static final double[] inputProbab    = {0.13 , 0.23, 0.64};
//   you need the CDF
    private static final double[] inputProbab    = {0.13 , 0.36, 1.0};
    
    
    // ------------------------------------------------------------------------
    // Fields
    // ------------------------------------------------------------------------

    /** Protocol identifier; obtained from config property {@link #PAR_PROT}. */
    private final int pid;
    private final int minchildren, maxchildren, maxlevels, numchunks;

    // ------------------------------------------------------------------------
    // Constructor
    // ------------------------------------------------------------------------

    /**
     * Creates a new instance and read parameters from the config file.
     */
    public TFDInitializer(String prefix) {
        pid = Configuration.getPid(prefix + "." + PAR_PROT);
        minchildren = Configuration.getInt(prefix + "." + PAR_MINCHILDREN, 1);
        maxchildren = Configuration.getInt(prefix + "." + PAR_MAXCHILDREN, 4);
        maxlevels   = Configuration.getInt(prefix + "." + PAR_MAXLEVELS,   1000);
        numchunks   = Configuration.getInt(prefix + "." + PAR_NUMCHUNKS,   100);

    }

    // ------------------------------------------------------------------------
    // Methods
    // ------------------------------------------------------------------------

    /**
    * Initialize an aggregation protocol using a peak distribution.
    * That is, one node will get the peek value, the others zero.
    * @return always false
    */
    public boolean execute() {
    
	int index, i, j;
	int MAXVALUE = 1000; 
	
	for (i = 0; i < Network.size(); i++) {
		TreeFileDistr prot = (TreeFileDistr) Network.get(i).getProtocol(pid);
		prot.resetAll();
		prot.setFileTime(100/inputBandwidth[0]);
		prot.setFileTimeTree(100/inputBandwidth[0]);
		index = CommonState.r.nextInt(MAXVALUE);
		prot.setBW(inputBandwidth[0]);
		for (j = 0; j < inputProbab.length-1; j++){
			if (index > MAXVALUE*inputProbab[j]){
				prot.setBW(inputBandwidth[j+1]);
			}//if
		}//for j
        }
        
	Node node = getPeer();
	TreeFileDistr prot = (TreeFileDistr) node.getProtocol(pid);
	/* set input parameters; if they are not present, default values will be used (1, 4 , 1000 and 100) */
	prot.setMinChildren(minchildren);
	prot.setMaxChildren(maxchildren);
	prot.setMaxLevels(maxlevels);
	prot.setNumChunks(numchunks);
	
	prot.setLevel(1);
	prot.setBW(inputBandwidth[inputBandwidth.length-1]);
	prot.setRate(inputBandwidth[inputBandwidth.length-1]/2);
	prot.setChunkTime(1/prot.getRate());
	prot.setFileTime(numchunks/prot.getRate());
	prot.setFileTimeTree(numchunks/prot.getRate());
	prot.setStripe(1);
	EDSimulator.add((long)prot.getChunkTime(), new ArrivedMessage(5, node, 1, 0), node, pid);
	/* the '5' stands for the event type (FIRSTCOMPL), and '1' for the stripe identifier*/
	
	/* get another node, with different stripe    */
	node = getPeer();
	prot = (TreeFileDistr) node.getProtocol(pid);
	while (prot.getLevel()>0) {
		node = getPeer();
		prot = (TreeFileDistr) node.getProtocol(pid);
	}
	prot.setMinChildren(minchildren);
	prot.setMaxChildren(maxchildren);
	prot.setMaxLevels(maxlevels);
	prot.setNumChunks(numchunks);
	
	prot.setLevel(1);
	prot.setBW(inputBandwidth[inputBandwidth.length-1]);
	prot.setRate(inputBandwidth[inputBandwidth.length-1]/2);
	prot.setChunkTime(1/prot.getRate());
	prot.setFileTime(numchunks/prot.getRate());
	prot.setFileTimeTree(numchunks/prot.getRate());
	prot.setStripe(2);
	EDSimulator.add((long)prot.getChunkTime(), new ArrivedMessage(5, node, 2, 0), node, pid);
	return false;
    }
    
    
    // ====================== helper methods ==============================
// ====================================================================

/**
 * Returns a peer node which is accessible (has ok fail state). This
 * implementation starts with a random node. If that is not reachable,
 * proceed first towards the older and then younger nodes until the first
 * reachable is found.
 * @return null if no accessible peers are found, the peer otherwise.
 */
private Node getPeer()
{

	final int d = Network.size();
	if (d == 0)
		return null;
	int index = CommonState.r.nextInt(d);
	Node result = Network.get(index);

	while (!result.isUp()){
		index = CommonState.r.nextInt(d);
		result = Network.get(index);
	}
	if(result.isUp())
		return result;

	// no accessible peer
	return null;
}

}
