/*
 * Copyright (c) 2003 The BISON Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package mycode.treefiledistr;

import peersim.config.*;
import peersim.core.*;
import peersim.transport.Transport;
import peersim.cdsim.CDProtocol;
import peersim.edsim.*;
import java.lang.Double;

/**
* Event driven version of epidemic averaging.
*/
public class TreeFileDistr implements CDProtocol, EDProtocol {

//--------------------------------------------------------------------------
// Initialization
//--------------------------------------------------------------------------

/**
 * @param prefix string prefix for config properties
 */
public TreeFileDistr(String prefix) {
}

//--------------------------------------------------------------------------
//Fields
//--------------------------------------------------------------------------

/* these constant are equal for all the nodes and fixed */
protected static final int BRANDNEW    = 0;
protected static final int REPEATEDNEW = 1;
protected static final int REPLYNOT    = 2;
protected static final int REPLYYES    = 3;
protected static final int GOONMSG     = 4;
protected static final int FIRSTCOMPL  = 5;
protected static final int RESETSTRIPE = 6;

protected static final int NOSTRIPE    = 0;
protected static final int STRIPE_A    = 1;
protected static final int STRIPE_B    = 2;
protected static final int STRIPE_BOTH = 3;


protected static final double PERCENTAGEBW   = 0.75;

/* these constant are equal for all the nodes and initialized at the beginning */
protected static int MINCHILDREN = 1;
protected static int MAXCHILDREN = 4;
protected static int MAXLEVELS   = 1000;
protected static int NUMCHUNKS   = 100;




/** Value held by this protocol */
protected  boolean reached;       /* it stores the number of level in the diffusion tree */
protected  int     level;        /* it stores the number of level in the diffusion tree */
protected  double  bw;           /* the download bandwidth of the node */
protected  double  rate;         /* the actual rate used to transfer chunks*/
protected  double  endChunk;     /* the instant when the first chunk is completely downloaded */
protected  double  endFile;      /* the instant when the whole file is completely downloaded */
protected  double  endFileTree;  /* the instant when the whole file is completely downloaded for a single tree */
protected  int     numChildren;  /* number of children the node has */
protected  int     stripe;       /* received stripe */

protected  double  totalChildrenRates, downloaded;
protected  boolean nodeSaturated, secondRound;

protected  int    numTries, receivedReplies, sentRequests;

private  double[]  childrenBw     = new double[20];
private  double[]  childrenRate   = new double[20];
private  int[]     childrenId     = new int[20];
private  int[]     childrenStripe = new int[20];
private  boolean[] childrenCompl  = new boolean[20];

//--------------------------------------------------------------------------
// methods
//--------------------------------------------------------------------------


/**
 * Clones the value holder.
 */
public Object clone()
{
	TreeFileDistr tfdh = null;
	try { tfdh = (TreeFileDistr)super.clone(); }
	catch( CloneNotSupportedException e ) {} // never happens
	tfdh.childrenBw   = new double[childrenBw.length];
	tfdh.childrenRate = new double[childrenRate.length];
	tfdh.childrenId   = new int[childrenId.length];
	tfdh.childrenStripe  = new int[childrenStripe.length];
	System.arraycopy(childrenBw,    0, tfdh.childrenBw,    0, childrenBw.length);
	System.arraycopy(childrenRate,  0, tfdh.childrenRate,  0, childrenRate.length);
	System.arraycopy(childrenId,    0, tfdh.childrenId,    0, childrenId.length);
	System.arraycopy(childrenStripe,0, tfdh.childrenStripe,0, childrenStripe.length);
	return tfdh;
}

public int getReached()
{
	if (reached) return(1);
	return (0);
}

public int getLevel()
{
	return level;
}

public double getBW()
{
	return bw;
}

public double getRate()
{
	return rate;
}

public double getSpareBw()
{
	return (bw-rate);
}


public double getChunkTime()
{
	return endChunk;
}

public double getFileTime()
{
	return endFile;
}

public double getFileTimeTree()
{
	return endFileTree;
}

public int getChildren()
{
	return numChildren;
}

public int getStripe()
{
	return stripe;
}
//--------------------------------------------------------------------------

/**
 * @inheritDoc
 */

public void resetAll()
{
	this.reached     = false;
	this.level       = 0;
	this.bw          = 0;
	this.rate        = 0;
	this.endChunk    = 0;
	this.endFile     = 0;
	this.endFileTree = 0;
	this.numChildren = 0;
	this.numTries    = 0;
	this.receivedReplies = 0;
	this.sentRequests = 0;
	this.downloaded = 0;
	this.totalChildrenRates = 0;
	this.nodeSaturated = false;
	this.secondRound = false;
	this.stripe = NOSTRIPE;
}
 

public void setMinChildren(int mchildren)
{
	this.MINCHILDREN = mchildren;
}

public void setMaxChildren(int children)
{
	this.MAXCHILDREN = children;
}

public void setMaxLevels(int mlevels)
{
	this.MAXLEVELS = mlevels;
}

public void setNumChunks(int nchunks)
{
	this.NUMCHUNKS = nchunks;
}

public void setReached(boolean status)
{
	this.reached = status;
}

 
public void setLevel(int level)
{
	this.level = level;
}

public void setBW(double bw)
{
	this.bw = bw;
}

public void setRate(double rate)
{
	this.rate = rate;
}

public void setChunkTime(double time)
{
	this.endChunk = time;
}

public void setFileTime(double time)
{
	this.endFile = time;
}

public void setFileTimeTree(double time)
{
	this.endFileTree = time;
}

public void setChildren(int children)
{
	this.numChildren = children;
}

public void setStripe(int stripe)
{
	this.stripe = stripe;
}

//--------------------------------------------------------------------------
public void nextCycle( Node node, int pid )
{
}

/**
* This is the standard method to define to process incoming messages.
*/
public void processEvent( Node node, int pid, Object event ) {
		
	Linkable linkable;
	Node peern;

	ArrivedMessage aem = (ArrivedMessage)event;
	
	switch(aem.typeOfMsg) {
		case RESETSTRIPE:
		/* after brandnew the node has decided not to send me the content, so
		   my level is resetted */
			stripe -= aem.stripeId;
			if (stripe == NOSTRIPE) 
				level = 0;
			break;
		case BRANDNEW:
		/* one of my neighbors informed me that he has brand new content (first chunk)
		   reply that I am not interested if I am already downloading 
		   reply "yes" if not yet reached and set the level accordingly */
			if (stripe != NOSTRIPE)
				((Transport)node.getProtocol(FastConfig.getTransport(pid))).
					send(
						node,
						aem.sender,
						new ArrivedMessage(REPLYNOT, node, aem.stripeId, 0),
						pid);
			else {
				stripe += aem.stripeId;
				level = (int)aem.msgParameter + 1;
				((Transport)node.getProtocol(FastConfig.getTransport(pid))).
					send(
						node,
						aem.sender,
						new ArrivedMessage(REPLYYES, node, aem.stripeId, bw-rate),
						pid);
			}
			break;
		case REPEATEDNEW:
		/* one of my neighbors informed me that he has brand new content (first chunk)
		   reply that I am not interested if I am already downloading 
		   reply "yes" if not yet reached and set the level accordingly */
			if ((stripe == STRIPE_BOTH)||(aem.stripeId == stripe))
				((Transport)node.getProtocol(FastConfig.getTransport(pid))).
					send(
						node,
						aem.sender,
						new ArrivedMessage(REPLYNOT, node, aem.stripeId, 0),
						pid);
			else {
				stripe += aem.stripeId;
				if (stripe != STRIPE_BOTH) 
					level = (int)aem.msgParameter + 1;
				((Transport)node.getProtocol(FastConfig.getTransport(pid))).
					send(
						node,
						aem.sender,
						new ArrivedMessage(REPLYYES, node, aem.stripeId, bw-rate),
						pid);
			}
			break;
		case REPLYNOT:
		/* one of the neighbor I contacted is not interested; 
		   if the number of receivedReplies is over sentRequests, process the children
		   otherwise wait */
			receivedReplies++;
			if ((receivedReplies == sentRequests)&&(numChildren > 0)&&(!nodeSaturated)){
				nodeSaturated = true;
				computeRates();
				for (int i = 0; i < numChildren; i++) {
					((Transport)node.getProtocol(FastConfig.getTransport(pid))).
					send(
						node,
						Network.get(childrenId[i]),
						new ArrivedMessage(GOONMSG, node, 
								childrenStripe[i], childrenRate[i]),
						pid);
				}
			}
			else if ((receivedReplies == sentRequests)&&(numChildren == 0)&&
				(!nodeSaturated)&&(!secondRound)){
				/* resend the requests, tagged ad "RepeatedNew" */
				secondRound = true;
				receivedReplies = 0;
				linkable = (Linkable) node.getProtocol( FastConfig.getLinkable(pid) );
				sentRequests = Math.min(4*MAXCHILDREN, linkable.degree());
				for (int k = 0; k < sentRequests; k++) {
					peern = linkable.getNeighbor(k);
					((Transport)node.getProtocol(FastConfig.getTransport(pid))).
						send(
							node,
							peern,
							new ArrivedMessage(REPEATEDNEW, node, aem.stripeId, (double)level),
							pid);
				}
			}
			break;
		case REPLYYES:
		/* check if all the nodes have replied, so you can go on, otherwise wait */
			receivedReplies++;
			if (nodeSaturated) {
			/* refuse to upload to previuosly announced nodes: reset their level */
				((Transport)node.getProtocol(FastConfig.getTransport(pid))).
					send(
						node,
						aem.sender,
						new ArrivedMessage(RESETSTRIPE, node, aem.stripeId, 0),
						pid);
			}
			else {
			/* add the new node to your children list */
				childrenId[numChildren] = aem.sender.getIndex();
				childrenBw[numChildren] = aem.msgParameter;
				childrenStripe[numChildren] = aem.stripeId;
				numChildren++;
				totalChildrenRates += Math.min(aem.msgParameter, rate);
				if (((totalChildrenRates > PERCENTAGEBW*bw)||(receivedReplies == sentRequests)||
					(numChildren==MAXCHILDREN))&&((numChildren>=MINCHILDREN))){
					nodeSaturated = true;
					computeRates();
					for (int i = 0; i < numChildren; i++) {
						((Transport)node.getProtocol(FastConfig.getTransport(pid))).
						send(
							node,
							Network.get(childrenId[i]),
							new ArrivedMessage(GOONMSG, node, 
									childrenStripe[i], childrenRate[i]),
							pid);
					}
				}
			}
			break;
		case GOONMSG:
		/* compute your end time and schedule when the first chunk will be completed */
			if (!reached){
				reached      = true;
				rate         = aem.msgParameter;
				endChunk     = CommonState.getTime() + 1/rate;
				endFileTree  = endChunk + (NUMCHUNKS-1)/rate;
				endFile      = endChunk + (NUMCHUNKS-1)/rate;
				Double delay = new Double(1.0/rate);
				EDSimulator.add(delay.longValue(), new ArrivedMessage(FIRSTCOMPL, node, aem.stripeId, 0), node, pid);
			}
			else {
				downloaded = (CommonState.getTime() - endChunk)*rate + 1; //+1 is for the first chunk
				rate      += Math.min(aem.msgParameter, bw-rate);
				endFile    = CommonState.getTime() + (NUMCHUNKS-downloaded)/rate;
			}
			break;
		case FIRSTCOMPL:
		/* I have finished to download the first chunk, I can start contacting nodes for
		   sending this brand new data; since initialization is random, I selects
		   nodes starting from the first neighbor and I send a number of requests
		   that is the double of the number of maximum children */
			if (level < MAXLEVELS){
				linkable = (Linkable) node.getProtocol( FastConfig.getLinkable(pid) );
				sentRequests = Math.min(4*MAXCHILDREN, linkable.degree());
				for (int k = 0; k < sentRequests; k++) {
					peern = linkable.getNeighbor(k);
					((Transport)node.getProtocol(FastConfig.getTransport(pid))).
						send(
							node,
							peern,
							new ArrivedMessage(BRANDNEW, node, aem.stripeId, (double)level),
							pid);
				}
			}
			break;
	}

}


// ====================== helper methods ==============================
// ====================================================================

/**
 * Compute the rates according to max-min fairness criterion
 */

private void computeRates () {

 double sumSpare, bwToBeDistributed;
 int    sumNotCompl, i, r;

	bwToBeDistributed = bw;
	/* reset vectors */
	for (i = 0; i < numChildren; i++){
		childrenRate[i] = 0.0; 
		childrenCompl[i] = false; 
	}//for i 
	
	sumNotCompl = numChildren;
	while(sumNotCompl>0) {   
		sumSpare = 0;
		for (r=0; r<numChildren; r++){
			if(!childrenCompl[r]){    
				childrenRate[r] += bwToBeDistributed/sumNotCompl;
			}//if
			if(childrenRate[r] > childrenBw[r]){
				childrenCompl[r] = true;
				sumSpare += childrenRate[r] - childrenBw[r];
				childrenRate[r] = childrenBw[r];    
			}//if
		}//for r
		if(sumSpare > 0){
			bwToBeDistributed = sumSpare;
			sumNotCompl = 0;
			for (r=0; r<numChildren; r++){
				if(!childrenCompl[r])
					sumNotCompl++;
			}//for r
		}//if
		else { sumNotCompl = 0; }    
	}//while   
	
	for (i = 0; i < numChildren; i++) {
		childrenRate[i] = Math.min(rate, childrenRate[i]);
	}
}


}

//--------------------------------------------------------------------------
//--------------------------------------------------------------------------

/**
* The type of a message. It contains a value of type double and the
* sender node of type {@link peersim.core.Node}.
*/
class ArrivedMessage {

	final int    typeOfMsg;
	final double msgParameter;
	final int    stripeId;
	final Node   sender;
	public ArrivedMessage(int typeOfMsg, Node sender, int stripeId, double msgParameter )
	{
		this.typeOfMsg    = typeOfMsg;
		this.sender       = sender;
		this.stripeId     = stripeId;
		this.msgParameter = msgParameter;
	}
}

